from .baseservice import BaseService

from .components import Activity_Report, Contact, Contact_List, Email_Address, Result_Set, Tracking_Summary

import sys
if sys.version_info[0] == 2:
    from string import join

class ContactService(BaseService):
    # Contacts services endpoints    
    def get_contacts(self, email = None, limit = 50, modified_since = None, status = 'ALL'):
        queries = {}
        if email is not None:
            if type(email) is Email_Address:
                email = email.get_email_address()
            queries['email'] = email
        
        queries['limit'] = limit
        queries['status'] = status

        if modified_since is not None:
            queries['modified_since'] = modified_since
        
        response = self.request('get', 'contacts', params = queries)
        try:
            response['results']
        except:
            return response
        # Convert dict to Components
        for i in range(len(response['results'])):
            response['results'][i] = Contact(response['results'][i])
        return Result_Set(response)

    def post_contacts(self, contact, action_by = 'ACTION_BY_OWNER'):
        contact = self.validate_contact(contact)
        return Contact(self.request('post', 'contacts', params = {'action_by': action_by}, body = contact))

    def get_contact(self, contact_id):
        contact_id = self.contact_to_id(contact_id)
        return Contact(self.request('get', 'contacts/'+contact_id))

    def update_contact(self, contact, action_by = 'ACTION_BY_OWNER'):
        contact = self.validate_contact(contact)
        if not contact.get_id():
            raise AttributeError('missing Contact.id')
        return Contact(self.request('put', 'contacts/'+contact.get_id(), params = {'action_by': action_by}, body = contact))

    def delete_contact(self, contact_id):
        contact_id = self.contact_to_id(contact_id)
        return self.request('delete', 'contacts/'+contact_id) 

    # Contacts tracking endpoints
    def get_contact_tracking(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking', params = queries)

        return self.tracking_result_set(response)

    def get_contact_tracking_summary(self, contact_id):
        contact_id = self.contact_to_id(contact_id)

        return Tracking_Summary(self.request('get', 'contacts/'+contact_id+'/tracking/reports/summary'))

    def get_contact_bounces(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/bounces', params = queries)

        return self.tracking_result_set(response)

    def get_contact_clicks(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/clicks', params = queries)

        return self.tracking_result_set(response)

    def get_contact_forwards(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/forwards', params = queries)

        return self.tracking_result_set(response)

    def get_contact_opens(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/opens', params = queries)

        return self.tracking_result_set(response)


    def get_contact_sends(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/sends', params = queries)

        return self.tracking_result_set(response)

    def get_contact_unsubscribes(self, contact_id, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)
        contact_id = self.contact_to_id(contact_id)

        response = self.request('get', 'contacts/'+contact_id+'/tracking/unsubscribes', params = queries)

        return self.tracking_result_set(response)

    def get_contact_report_by_campaign(self, contact_id):
        contact_id = self.contact_to_id(contact_id)

        response = {'results': self.request('get', 'contacts/'+contact_id+'/tracking/reports/summaryByCampaign')}
        return self.tracking_result_set(response)

    def validate_contact(self, contact):
        if type(contact) is dict:
            contact = Contact(contact)
        if type(contact) is not Contact:
            raise TypeError('invalid type for contact: '+type(contact).__name__)

        return contact

    def contact_to_id(self, contact_id):
        if type(contact_id) is Contact:
            contact_id = contact_id.get_id()
        return contact_id

    def tracking_queries(self, created_since, limit):
        queries = {}

        if created_since is not None:
            queries['created_since'] = created_since
        
        queries['limit'] = limit

        return queries

    def tracking_result_set(self, tracking_response):
        try:
            tracking_response['results']
        except:
            return tracking_response
        for i in range(len(tracking_response['results'])):
            tracking_response['results'][i] = Activity_Report(tracking_response['results'][i])
        return Result_Set(tracking_response)



from .domain_lambda import *

from .contact import Contact
from .account_info import Account_Info

class Event(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'active_date',
            'are_registrants_public',
            'cancelled_date',
            'created_date',
            'currency_type',
            'deleted_date',
            'description',
            'end_date',
            'google_analytics_key',
            'google_merchant_id',
            'id',
            'is_calendar_displayed',
            'is_checkin_available',
            'is_home_page_displayed',
            'is_listed_in_external_directory',
            'is_map_displayed',
            'is_virtual_event',
            'location',
            'meta_data_tags',
            'name',
            'payable_to',
            'paypal_account_email',
            'registration_url',
            'start_date',
            'status',
            'theme_name',
            'time_zone_id',
            'title',
            'total_registered_count',
            'type',
            'twitter_hash_tag',
            'updated_date'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')

    def get_contact(self):
        try:
            return self['contact']
        except:
            return None

    def set_contact(self, contact):
        if type(contact) is dict:
            self['contact'] = contact
        # Prune all other fields for a event contact
        elif type(contact) is Contact:
            event_contact = {}
            try:
                event_contact['email_address'] = contact.get_email_address().get_email_address()
            except:
                pass
            # Automatically generate name
            name = ''
            name = name +(contact.get_prefix_name() if contact.get_prefix_name() else '')
            name = name +((' '+ contact.get_first_name()) if contact.get_first_name() else '')
            name = name +((' '+ contact.get_last_name()) if contact.get_last_name() else '')
            event_contact['name'] = name
            event_contact['organization_name'] = contact.get_company_name()
            event_contact['phone_number'] = contact.get_work_phone()
            self['contact'] = event_contact
        elif type(contact) is Account_Info:
            event_contact = {}
            event_contact['email_address'] = contact.get_email()
            # Automatically generate name
            name = ''
            name = name +(contact.get_first_name() if contact.get_first_name() else '')
            name = name +((' '+ contact.get_last_name()) if contact.get_last_name() else '')
            event_contact['name'] = name
            event_contact['organization_name'] = contact.get_organization_name()
            event_contact['phone_number'] = contact.get_phone()
            self['contact'] = event_contact

    def delete_contact(self):
        try:
            self.pop('contact', None)
        except:
            pass



from .domain_lambda import *

# Multipart File component
class Multipart_File(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'data',
            'description',
            'file_name',
            'file_type',
            'folder_id',
            'source'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            


from .domain_lambda import *

class Contact_List(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'contact_count',
            'created_date',
            'id',
            'modified_date',
            'name',
            'status'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
        self.set_name = make_setter(self, 'name')
        self.set_status = make_setter(self, 'status')



from .baseservice import BaseService

from .components import Activity_Status, Contact, Email_Address, Multipart_Activity

class ActivityService(BaseService):
    # Activity services endpoints
    def import_contacts(self, import_data, lists, columns):
        if 'EMAIL' in columns:
            for contact in import_data:
                if type(contact) is Contact:
                    for i in range(len(contact['email_addresses'])):
                        if type(contact['email_addresses'][i]) is dict or type(contact['email_addresses'][i]) is Email_Address:
                            contact['email_addresses'][i] = contact['email_addresses'][i]['email_address']
                    
        body = {'import_data': import_data, 'lists': lists, 'column_names': columns}
        return Activity_Status(self.request('post', 'activities/addcontacts', body = body))

    def remove_contacts(self, import_data, lists):
        for contact in import_data:
            if type(contact) is Contact:
                for i in range(len(contact['email_addresses'])):
                    if type(contact['email_addresses'][i]) is dict or type(contact['email_addresses'][i]) is Email_Address:
                        contact['email_addresses'][i] = contact['email_addresses'][i]['email_address']

        body = {'import_data': import_data, 'lists': lists}
        return Activity_Status(self.request('post', 'activities/removefromlists', body = body))

    def clear_contact_lists(self, lists):
        return Activity_Status(self.request('post', 'activities/clearlists', body = {'lists': lists}))

    def export_contacts(self, columns, lists, sort_by = 'EMAIL_ADDRESS', file_type = 'CSV', export_added_by = False, export_date_added = False):
        # The body is constructed from individual components passed in by the developer
        body = {'lists': lists, 'column_names': columns, 'file_type': file_type, 'sort_by': sort_by, 'export_added_by': export_added_by, 'export_date_added': export_date_added}
        return Activity_Status(self.request('post', 'activities/exportcontacts', body = body))

    # Retrieves the .csv data
    def get_export_file(self, file_name):
        file_name = file_name.split('v2/')[1].split('?')[0]
        return self.request('get', file_name)

    def multipart_import_contacts(self, multipart_activity):
        attributes = [
            'data',
            'file_name',
            'lists'
        ]
        form_file, form_data = self.prepare_multipart_body(multipart_activity, attributes)

        return self.request('post', 'activities/addcontacts', form_file = form_file, form_data = form_data)

    def multipart_remove_contacts(self, multipart_activity):
        attributes = [
            'data',
            'file_name',
            'lists'
        ]
        form_file, form_data = self.prepare_multipart_body(multipart_activity, attributes)

        return self.request('post', 'activities/removefromlists', form_file = form_file, form_data = form_data)

    def activities_status_report(self, type_query = None, status = 'ALL'):
        params = {'status': status}
        if type_query:
            params['type'] = type_query
        
        response = self.request('get', 'activities', params = params)
        
        if type(response) is list:
            for i in range(len(response)):
                response[i] = Activity_Status(response[i])
        return response

    def individual_activity_status(self, activity_id):
        activity_id = self.activity_to_id(activity_id)
        return Activity_Status(self.request('get', 'activities/'+activity_id))

    def activity_to_id(self, activity_id):
        if type(activity_id) is Activity_Status:
            activity_id = activity_id.get_id()
        return activity_id



from .account_info import Account_Info
from .activity_report import Activity_Report
from .activity_status import Activity_Status
from .address import Address
from .attribute import Attribute
from .billing_plan import Billing_Plan
from .campaign import Campaign
from .contact import Contact
from .contact_list import Contact_List
from .email_address import Email_Address
from .event import Event
from .fee import Fee
from .item import Item
from .multipart_activity import Multipart_Activity
from .multipart_file import Multipart_File
from .mylibrary_file import Mylibrary_File
from .mylibrary_folder import Mylibrary_Folder
from .partner_account import Partner_Account
from .promocode import Promocode
from .registrant import Registrant
from .result_set import Result_Set
from .schedule import Schedule
from .tracking_summary import Tracking_Summary



from .domain_lambda import *

# Mylibrary File component
class Mylibrary_File(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Thumbnail is returned as a dict with
        # 3 properties: width, url, height
        attributes = [
            'created_date',
            'description',
            'file_type',
            'folder',
            'folder_id',
            'height',
            'id',
            'is_image',
            'modified_date',
            'name',
            'size',
            'source',
            'status',
            'thumbnail',
            'url',
            'width'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            


# Lambda functions used to dynamically generate getter and setter methods
# These methods are applicable to dictionaries

# self refers to the object implementing these methods
# name refers to the field being modified
def make_getter(self, name):
    return lambda: get_attr(self, name)
def get_attr(self, name):
    try:
        return self[name]
    except:
        return None

def make_setter(self, name):
    # v refers to the value the field is being set to
    return lambda v: set_attr(self, name, v)
def set_attr(self, name, value):
    self[name] = value

def make_deleter(self, name):
    return lambda: delete_attr(self, name)
def delete_attr(self, name):
    try:
        self.pop(name, None)
    except:
        pass



from .domain_lambda import *

# Tracking_Summary is a simple component with no additional logic
class Tracking_Summary(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'bounces',
            'clicks',
            'forwards',
            'opens',
            'sends',
            'spam_count',
            'unsubscribes'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            


from .domain_lambda import *
from .address import Address

# Account_Info is a mostly simple component
class Account_Info(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'company_logo',
            'country_code',
            'email',
            'first_name',
            'last_name',
            'organization_name',
            'phone',
            'state_code',
            'time_zone',
            'website'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')

    # API currently only supports 1 organization address
    def get_organization_address(self):
        try: 
            self['organization_addresses'][0]
        except:
            return None
        if type(self['organization_addresses'][0]) is not Address:
            self['organization_addresses'][0] = Address(self['organization_addresses'][0])
        return self['organization_addresses'][0]
    # addresses for organization are different from addresses for contacts;
    # however, the model is the same so it is up to the user to adjust fields accordingly
    def set_organization_address(self, address):
        if type(address) is dict:
            address = Address(address)
        if type(address) is not Address:
            raise TypeError('invalid type for address: '+type(address).__name__)
        self['organization_addresses'] = [address]
    def clear_organization_address(self):
        try:
            self.pop('organization_addresses', None)
        except:
            pass



class Result_Set(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        try:
            self.set_type = type(self['results'][0])
        except:
            self.set_type = None
    def get_set(self):
        return self['results']
    def get_item(self, index):
        return self['results'][index]
    def get_set_type(self):
        return self.set_type
    def add_item(self, item):
        if self.set_type:
            if type(item) is not self.set_type:
                raise TypeError('Item type ('+type(item).__name__+') does not match set_type ('+set_type.__name__+')')
            else:
                self['results'].append(item)
        else:
            self['results'] = [item]
            self.set_type = type(item)
    def remove_item(self, index):
        self['results'].pop(index)
        if not self['results']:
            self.set_type = None
    def clear_result_set(self):
        self.pop('results', None)
        self['results'] = []
        self.set_type = None
    def get_next_page(self):
        try:
            return self['meta']['pagination']['next_link']
        except:
            return None



from .components import Account_Info
from .components import Activity_Report
from .components import Activity_Status
from .components import Address
from .components import Attribute
from .components import Billing_Plan
from .components import Campaign
from .components import Contact
from .components import Contact_List
from .components import Email_Address
from .components import Event
from .components import Fee
from .components import Item
from .components import Multipart_Activity
from .components import Multipart_File
from .components import Mylibrary_File
from .components import Mylibrary_Folder
from .components import Partner_Account
from .components import Promocode
from .components import Registrant
from .components import Result_Set
from .components import Schedule
from .components import Tracking_Summary

from .constantcontact import ConstantContact

from .accountservice import AccountService
from .activityservice import ActivityService
from .contactservice import ContactService
from .emailmarketingservice import EmailMarketing
from .eventspotservice import EventSpotService
from .listservice import ListService
from .mylibraryservice import MyLibraryService
from .partnerservice import PartnerService

from .baseservice import __version__



import requests
import json

from .components import Activity_Report, Campaign, Contact, Result_Set

__version__ = '0.3.4'

class BaseService():
    def __init__(self, client_id, token, base_url = 'https://api.constantcontact.com/v2'):
        self.client_id = client_id
        self.token = token
        # Add required headers.
        # TODO: Add header for SDK usage tracking
        self.header = {}
        self.header['Authorization'] = 'Bearer '+self.token
        self.header['x-ctct-request-source'] = 'sdk.lwang_python.'+__version__
        self.base_url = base_url

    # Token info only works on production environments. 
    def token_info(self):
        token_info = requests.post('https://oauth2.constantcontact.com/oauth2/tokeninfo.htm'+'?access_token='+self.token)
        token_json = json.loads(token_info.content)
        return token_json

    def request(self, request_type, endpoint, params = {}, body = [], header = {}, form_data = None, form_file = None):
        # Formulate a request URI string for a request
        request_string = self.base_url+'/'+endpoint
        
        params['api_key'] = self.client_id

        response = None

        # set required headers.
        header.update(self.header)

        if request_type == 'get':
            response = requests.get(request_string, headers=header, json=body, params = params)
        elif request_type == 'put':
            response = requests.put(request_string, headers=header, json=body, params = params)
        elif request_type == 'post':
            # For multipart/form-data posts (Mylibrary multipart post)
            if form_data and form_file:
                response = requests.post(request_string, headers=header, params = params, files = form_file, data = form_data)
            else:
                response = requests.post(request_string, headers=header, json=body, params = params)
        elif request_type == 'delete':
            response = requests.delete(request_string, headers=header, json=body, params = params)
        elif request_type == 'patch':
            response = requests.patch(request_string, headers=header, json=body, params = params)


        try:
            response = response if response is None else json.loads(response.content.decode('utf8'))
        except:
            pass
        '''
        except AttributeError:
            return response
        except ValueError:
            return response
        except TypeError:
        '''

        # Check if an error message key pair is returned
        try:
            if response[0]['error_message']:
                response = response[0]
        except:
            pass

        return response

    # The url from the pagination data can be directly passed into this method
    # Pagination is returned from the api in the form /v2/{service}?next={uuid}
    def next_page(self, url):
        service = url.split('/', 2)[2].split('?', 1)[0]
        url = self.base_url + '/' + url.split('/', 2)[2]
        
        # Set the api key for pagination request
        params = {'api_key': self.client_id}

        response = requests.get(url, headers=self.header, params = params)

        try:
            response = response if response is None else json.loads(response.content.decode('utf8'))
        except:
            pass
        # Check if an error message key pair is returned
        try:
            if response[0]['error_message']:
                response = response[0]
        except:
            # Parse response into a resultset
            if service == 'contacts':
                for i in range(len(response['results'])):
                    response['results'][i] = Contact(response['results'][i])
            elif service == 'emailmarketing/campaigns':
                for i in range(len(response['results'])):
                    response['results'][i] = Campaign(response['results'][i])
            elif service.split('/', 1)[0] == 'lists':
                for i in range(len(response['results'])):
                    response['results'][i] = Contact(response['results'][i])
            elif service.split('/')[2] == 'tracking':
                # Contact tracking
                for i in range(len(response['results'])):
                    response['results'][i] = Activity_Report(response['results'][i])
            elif service.split('/')[3] == 'tracking':
                # Email campaign tracking
                for i in range(len(response['results'])):
                    response['results'][i] = Activity_Report(response['results'][i])

        return Result_Set(response)
    
    # Prepares a multipart object into form_file and form_data into 
    # multipart/form-data content for a multipart post endpoint.
    # Returns a tuple containing the file and data fields.
    def prepare_multipart_body(self, multipart_object, attributes):
        form_file = {}
        form_data = {}

        for key, value in multipart_object.items():
            if key not in attributes:
                raise AttributeError('Invalid attribute for '+type(multipart_object).__name__+': '+key)
            if key == 'data':
                form_file[key] = value
            else:
                form_data[key] = (None, value)
        for key in attributes:
            try:
                multipart_object[key]
            except:
                raise AttributeError('Missing attribute for '+type(multipart_object).__name__+': '+key)

        return (form_file, form_data)




from .domain_lambda import *

# Fee is a simple component with no additional logic
class Fee(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'early_fee',
            'fee',
            'fee_scope',
            'has_restricted_access',
            'id',
            'label',
            'late_fee'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')


from .domain_lambda import *

# Partner_Account is a simple component with no additional logic
class Partner_Account(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'billing_locale',
            'country_code',
            'email',
            'first_name',
            'id',
            'idp_provider',
            'idp_provider_id',
            'last_name',
            'oauth2_token',
            'last_login_date',
            'organization_name',
            'password',
            'phone',
            'state_code',
            'status',
            'username'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')


from .baseservice import BaseService

from .components import Contact, Contact_List, Result_Set

class ListService(BaseService):
    # Contact List services endpoints
    # returns a list rather than a Result_Set
    def get_lists(self, modified_since = None):
        queries = {}
        if modified_since:
            queries['modified_since'] = modified_since

        response = self.request('get', 'lists', params = queries)
        # Convert dict to components
        if type(response) is list:
            for i in range(len(response)):
                response[i] = Contact_List(response[i])
        return response

    def post_lists(self, contact_list, status = 'ACTIVE'):
        if type(contact_list) is str:
            contact_list = Contact_List({'name': contact_list, 'status': status})
        contact_list = self.validate_contact_list(contact_list)
        return Contact_List(self.request('post', 'lists', body = contact_list))

    # Methods which accept a list_id will also accept a list component
    def get_list(self, list_id):
        list_id = self.list_to_id(list_id)
        return Contact_List(self.request('get', 'lists/'+list_id))

    def update_list(self, contact_list):
        contact_list = self.validate_contact_list(contact_list)
        return Contact_List(self.request('put', 'lists/'+contact_list.get_id(), body = contact_list))

    def delete_list(self, list_id):
        list_id = self.list_to_id(list_id)
        return self.request('delete', 'lists/'+list_id)

    # TODO: This method does not currently return a Result_Set of Contact components
    def get_list_contacts(self, list_id, limit = 50, modified_since = None):
        queries = {}
        queries['limit'] = limit
        if modified_since is not None:
            queries['modified_since'] = modified_since

        list_id = self.list_to_id(list_id)
        response = self.request('get', 'lists/'+list_id+'/contacts', params = queries)

        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Contact(response['results'][i])
        return Result_Set(response)

    def validate_contact_list(self, contact_list):
        if type(contact_list) is dict:
            contact_list = Contact_List(contact_list)
        if type(contact_list) is not Contact_List:
            TypeError('invalid type for contact_list: '+type(contact_list).__name__)
            
        return contact_list

    def list_to_id(self, contact_list):
        if type(contact_list) is Contact_List:
            contact_list = contact_list.get_id()
        return contact_list



from .baseservice import BaseService

from .components import Activity_Report, Campaign, Email_Address, Result_Set, Schedule, Tracking_Summary

class EmailMarketing(BaseService):

    # Email Marketing Endpoints
    def get_campaigns(self, limit = 50, modified_since = None, status = 'ALL'):
        queries = {}

        if modified_since is not None:
            queries['modified_since'] = modified_since

        queries['status'] = status
        queries['limit'] = limit

        response = self.request('get', 'emailmarketing/campaigns', params = queries)
        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Campaign(response['results'][i])
        return Result_Set(response)

    def post_campaigns(self, campaign):
        campaign = self.validate_campaign(campaign)
        return Campaign(self.request('post', 'emailmarketing/campaigns', body = campaign))

    # Methods that are passed in a campaign_id will also accept a Campaign component.
    def get_campaign(self, campaign_id, updateSummary = False):
        queries = {}
        if updateSummary:
            queries[updateSummary] = 'true'

        campaign_id = self.campaign_to_id(campaign_id)
        return Campaign(self.request('get', 'emailmarketing/campaigns/'+campaign_id, params = queries))

    def update_campaign(self, campaign):
        campaign = self.validate_campaign(campaign)
        if not campaign.get_id():
            raise AttributeError('missing Campaign.id')
        return Campaign(self.request('put', 'emailmarketing/campaigns/'+campaign.get_id(), body = campaign))

    def delete_campaign(self, campaign_id):
        campaign_id = self.campaign_to_id(campaign_id)
        return self.request('delete', 'emailmarketing/campaigns/'+campaign_id)

    # Email Marketing Testing Endpoints
    # Accepts an array of strings or an array of Email_Addresses for email_addresses
    # Accepts strings for format and personal_message
    def test_send_campaign(self, campaign_id, email_addresses, format, personal_message):
        test_send_body = {}
        # Convert Email_Address to string
        for i in range(len(email_addresses)):
            if type(email_addresses[i]) is Email_Address:
                email_addresses[i] = email_addresses[i].get_email_address()

        test_send_body['email_addresses'] = email_addresses
        test_send_body['format'] = format.upper()
        test_send_body['personal_message'] = personal_message
        
        campaign_id = self.campaign_to_id(campaign_id)
        return self.request('post', 'emailmarketing/campaigns/'+campaign_id+'/tests', body = test_send_body)

    def preview_campaign(self, campaign_id):
        campaign_id = self.campaign_to_id(campaign_id)
        return self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/preview')

    # Email Marketing Scheduling Endpoints
    # Returns a list of schedules rather than a Result_Set
    def get_campaign_schedules(self, campaign_id):
        campaign_id = self.campaign_to_id(campaign_id)
        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/schedules')
        try:
            response['results']
        except:
            return response
        for i in range(len(response)):
            response[i] = Schedule(response[i])
        return response
    # Accepts an ISO-8601 date as a string or a Schedule component
    def post_campaign_schedules(self, campaign_id, schedule):
        campaign_id = self.campaign_to_id(campaign_id)
        if type(schedule) is str:
            schedule = {'scheduled_date': schedule}

        response = self.request('post', 'emailmarketing/campaigns/'+campaign_id+'/schedules', body = schedule)
        return Schedule(response)

    def get_campaign_schedule(self, campaign_id, schedule_id):
        campaign_id = self.campaign_to_id(campaign_id)
        schedule_id = self.schedule_to_id(schedule_id)
        return Schedule(self.request('get', 'emailmarketing/campaigns/' + campaign_id + '/schedules/' + schedule_id))

    def update_campaign_schedule(self, campaign_id, schedule):
        campaign_id = self.campaign_to_id(campaign_id)
        schedule = self.validate_schedule(schedule)
        return Schedule(self.request('put', 'emailmarketing/campaigns/'+campaign_id+'/schedules/'+schedule.get_id(), body = schedule))

    def delete_campaign_schedule(self, campaign_id, schedule_id):
        campaign_id = self.campaign_to_id(campaign_id)
        schedule_id = self.schedule_to_id(schedule_id)
        return self.request('delete', 'emailmarketing/campaigns/'+campaign_id+'/schedules/'+schedule_id)

    # Email Marketing Tracking Endpoints
    def get_campaign_summary_report(self, campaign_id, updateSummary = False):
        if updateSummary:
            queries = {'updateSummary': 'true'}
        else:
            queries = {'updateSummary': 'false'}
        campaign_id = self.campaign_to_id(campaign_id)
        return Tracking_Summary(self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/reports/summary', params = queries))
            
    def get_campaign_bounces(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)

        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/bounces', params = queries)
        
        return self.tracking_result_set(response)

    def get_campaign_clicks(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)

        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/clicks', params = queries)
        
        return self.tracking_result_set(response)

    def get_campaign_forwards(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)
        
        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/forwards', params = queries)
        
        return self.tracking_result_set(response)

    def get_campaign_opens(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)
        
        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id +'/tracking/opens', params = queries)
        
        return self.tracking_result_set(response)

    def get_campaign_sends(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)
        
        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/sends', params = queries)
        
        return self.tracking_result_set(response)

    def get_campaign_unsubscribes(self, campaign_id, created_since = None, limit = 500):
        campaign_id = self.campaign_to_id(campaign_id)
        queries = self.tracking_queries(created_since, limit)
        
        response = self.request('get', 'emailmarketing/campaigns/'+campaign_id+'/tracking/unsubscribes', params = queries)

        return self.tracking_result_set(response)

    def get_campaign_clicks_by_link_id(self, campaign_id, link_id = None, created_since = None, limit = 500):
        queries = self.tracking_queries(created_since, limit)

        campaign_id = self.campaign_to_id(campaign_id)
        url = 'emailmarketing/campaigns/'+campaign_id+'/tracking/clicks'
        if link_id:
            url = url + '/' + link_id
        response = self.request('get', url, params = queries)

        return self.tracking_result_set(response)

    def campaign_to_id(self, campaign):
        if type(campaign) is Campaign:
            campaign = campaign.get_id()
        return campaign

    def schedule_to_id(self, schedule):
        if type(schedule) is Schedule:
            schedule = schedule.get_id()
        return schedule

    def validate_campaign(self, campaign):
        if type(campaign) is dict:
            campaign = Campaign(campaign)
        if type(campaign) is not Campaign:
            raise TypeError('invalid type for campaign: '+type(campaign).__name__)
        return campaign

    def validate_schedule(self, schedule):
        if type(schedule) is dict:
            schedule = Schedule(schedule)
        if type(schedule) is not Schedule:
            raise TypeError('invalid type for schedule: '+type(schedule).__name__)
        return schedule

    def tracking_queries(self, created_since, limit):
        queries = {}

        if created_since is not None:
            queries['created_since'] = created_since
        
        queries['limit'] = limit

        return queries

    def tracking_result_set(self, tracking_response):
        try:
            tracking_response['results']
        except:
            return tracking_response
        for i in range(len(tracking_response['results'])):
            tracking_response['results'][i] = Activity_Report(tracking_response['results'][i])
        return Result_Set(tracking_response)





from .domain_lambda import *

# Address is a simple component with no additional logic
class Address(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'id',
            'address_type',
            'latitude',
            'line1',
            'line2',
            'line3',
            'longitude',
            'city',
            'state_code',
            'state',
            'country_code',
            'postal_code',
            'sub_postal_code'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            


from .domain_lambda import *

from .contact_list import Contact_List
from .tracking_summary import Tracking_Summary

class Campaign(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'created_date',
            'email_content',
            'email_content_format',
            'from_email',
            'from_name',
            'greeting_name',
            'greeting_salutations',
            'greeting_string',
            'id',
            'is_permission_reminder_enabled',
            'is_view_as_webpage_enabled',
            'last_run_date',
            'message_footer',
            'modified_date',
            'name',
            'next_run_date',
            'permalink_url',
            'permission_reminder_text',
            'reply_to_email',
            'status',
            'style_sheet',
            'subject',
            'template_type',
            'text_content',
            'view_as_web_page_link_text',
            'view_as_web_page_text'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
    
    def get_click_through_details(self):
        try:
            return self['click_through_details']
        except:
            return None
    def get_click_through_detail(self, index):
        return self['click_through_details'][index]
    def clear_click_through_details(self):
        try:
            self.pop('click_through_details', None)
        except:
            pass
    
    # sent_to_contact_lists will not use the contact_list component, it 
    # will be a simple array of strings. The component will be passable 
    # to these methods, but only the list_id will be extracted and used.
    def get_sent_to_contact_lists(self):
        try:
            return self['sent_to_contact_lists']
        except:
            return None
    def add_sent_to_contact_lists(self, list_id):
        if type(list_id) is Contact_List:
            list_id = list_id.get_id()
        try:
            self['sent_to_contact_lists']
        except:
            self['sent_to_contact_lists'] = []
        self['sent_to_contact_lists'].append({'id': list_id})
    def remove_sent_to_contact_lists_id(self, list_id):
        try:
            self['sent_to_contact_lists']
        except:
            return None
        if type(list_id) is Contact_List:
            list_id = list_id.get_id()
        for i in range(len(self['sent_to_contact_lists'])):
            if self['sent_to_contact_lists'][i]['id'] == list_id:
                self['sent_to_contact_lists'].pop(i)
                break
        if len(self['sent_to_contact_lists']) == 0:
            self.clear_sent_to_contact_lists()
    def remove_sent_to_contact_lists(self, index):
        try:
            self['sent_to_contact_lists']
        except:
            return None
        self['sent_to_contact_lists'].pop(index)
        if len(self['sent_to_contact_lists']) == 0:
            self.clear_sent_to_contact_lists()
    def clear_sent_to_contact_lists(self):
        try:
            self.pop('sent_to_contact_lists', None)
        except:
            pass

    def get_tracking_summary(self):
        try:
            self['tracking_summary']
        except:
            return None
        if type(self['tracking_summary']) is dict:
            self['tracking_summary'] = Tracking_Summary(self['tracking_summary'])
        return self['tracking_summary']
    def delete_tracking_summary(self):
        try:
            self.pop('tracking_summary', None)
        except:
            pass




from .baseservice import BaseService

from .components import Attribute, Event, Fee, Item, Promocode, Registrant, Result_Set

class EventSpotService(BaseService):
    # EventSpot endpoints
    # Events Collection
    def get_events(self, limit = 50):
        queries = {}
        queries['limit'] = limit

        response = self.request('get', 'eventspot/events', params = queries)
        try:
            response['results']
        except:
            return response
        # Convert dict to Components
        for i in range(len(response['results'])):
            response['results'][i] = Event(response['results'][i])
        return Result_Set(response)
    def post_events(self, event):
        event = self.validate_event(event)
        return Event(self.request('post', 'eventspot/events', body = event))

    # Individual Event
    def get_event(self, event_id):
        event_id = self.event_to_id(event_id)
        return Event(self.request('get', 'eventspot/events/'+event_id))
    def update_event(self, event):
        event = self.validate_event(event)
        return Event(self.request('put', 'eventspot/events/'+event.get_id(), body = event))
    # Currently, only the REPLACE operation is supported and the status path value is allowed
    def patch_event(self, event_id, value):
        event_id = self.event_to_id(event_id)
        patch = [{'op': 'REPLACE', 'path': '#/status', 'value': value}]
        return Event(self.request('patch', 'eventspot/events/'+event_id, body = patch))
    
    # Event Fees
    # Returns a list of fees rather than a resultset of fees
    def get_event_fees(self, event_id):
        event_id = self.event_to_id(event_id)
        response = self.request('get', 'eventspot/events/'+event_id+'/fees')
        for i in range(len(response)):
            response[i] = Fee(response[i])
        return response
    def post_event_fees(self, event_id, fee):
        event_id = self.event_to_id(event_id)
        fee = self.validate_fee(fee)
        return Fee(self.request('post', 'eventspot/events/'+event_id+'/fees', body = fee))
    def get_event_fee(self, event_id, fee_id):
        event_id = self.event_to_id(event_id)
        fee_id = self.fee_to_id(fee_id)
        return Fee(self.request('get', 'eventspot/events/'+event_id+'/fees/'+fee_id))
    def update_event_fee(self, event_id, fee):
        event_id = self.event_to_id(event_id)
        fee = self.validate_fee(fee)
        return Fee(self.request('put', 'eventspot/events/'+event_id+'/fees/'+fee.get_id(), body = fee))
    def delete_event_fee(self, event_id, fee_id):
        event_id = self.event_to_id(event_id)
        fee_id = self.fee_to_id(fee_id)
        return self.request('delete', 'eventspot/events/'+event_id+'/fees/'+fee_id)
    
    def get_event_promocodes(self, event_id):
        event_id = self.event_to_id(event_id)
        response = self.request('get', 'eventspot/events/'+event_id+'/promocodes')
        for i in range(len(response)):
            response[i] = Promocode(response[i])
        return response
    def post_event_promocodes(self, event_id, promocode):
        event_id = self.event_to_id(event_id)
        promocode = self.validate_promocode(promocode)
        return Promocode(self.request('post', 'eventspot/events/'+event_id+'/promocodes', body = promocode))
    def get_event_promocode(self, event_id, promocode_id):
        event_id = self.event_to_id(event_id)
        promocode_id = self.promocode_to_id(promocode_id)
        return Promocode(self.request('get', 'eventspot/events/'+event_id+'/promocodes/'+promocode_id))
    def update_event_promocode(self, event_id, promocode):
        event_id = self.event_to_id(event_id)
        promocode = self.validate_promocode(promocode)
        return Promocode(self.request('put', 'eventspot/events/'+event_id+'/promocodes/'+promocode.get_id(), body = promocode))
    def delete_event_promocode(self, event_id, promocode_id):
        event_id = self.event_to_id(event_id)
        promocode_id = self.promocode_to_id(promocode_id)
        return self.request('delete', 'eventspot/events/'+event_id+'/promocodes/'+promocode_id)

    def get_event_items(self, event_id):
        event_id = self.event_to_id(event_id)
        response = self.request('get', 'eventspot/events/'+event_id+'/items')
        for i in range(len(response)):
            response[i] = Item(response[i])
        return response
    def post_event_items(self, event_id, item):
        event_id = self.event_to_id(event_id)
        item = self.validate_item(item)
        return Item(self.request('post', 'eventspot/events/'+event_id+'/items', body = item))
    def get_event_item(self, event_id, item_id):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        return Item(self.request('get', 'eventspot/events/'+event_id+'/items/'+item_id))
    def update_event_item(self, event_id, item):
        event_id = self.event_to_id(event_id)
        item = self.validate_item(item)
        return Item(self.request('put', 'eventspot/events/'+event_id+'/items/'+item.get_id(), body = item))
    def delete_event_item(self, event_id, item_id):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        return self.request('delete', 'eventspot/events/'+event_id+'/items/'+item_id)
    
    def get_event_item_attributes(self, event_id, item_id):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        response = self.request('get', 'eventspot/events/'+event_id+'/items/'+item_id+'/attributes')
        for i in range(len(response)):
            response[i] = Attribute(response[i])
        return response
    def post_event_item_attributes(self, event_id, item_id, attribute):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        attribute = self.validate_attribute(attribute)
        return Attribute(self.request('post', 'eventspot/events/'+event_id+'/items/'+item_id+'/attributes', body = attribute))
    def get_event_item_attribute(self, event_id, item_id, attribute_id):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        attribute_id = self.attribute_to_id(attribute_id)
        return Attribute(self.request('get', 'eventspot/events/'+event_id+'/items/'+item_id+'/attributes/'+attribute_id))
    def update_event_item_attribute(self, event_id, item_id, attribute):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        attribute = self.validate_attribute(attribute)
        return Attribute(self.request('put', 'eventspot/events/'+event_id+'/items/'+item_id+'/attributes/'+attribute.get_id(), body = attribute))
    def delete_event_item_attribute(self, event_id, item_id, attribute_id):
        event_id = self.event_to_id(event_id)
        item_id = self.item_to_id(item_id)
        attribute_id = self.attribute_to_id(attribute_id)
        return self.request('delete', 'eventspot/events/'+event_id+'/items/'+item_id+'/attributes/'+attribute_id)
    
    def get_event_registrants(self, event_id, limit = 50):
        queries = {}
        queries['limit'] = limit

        event_id = self.event_to_id(event_id)

        response = self.request('get', 'eventspot/events/'+event_id+'/registrants', params = queries)
        # Convert dict to Components
        for i in range(len(response['results'])):
            response['results'][i] = Registrant(response['results'][i])
        return Result_Set(response)

    def get_event_registrant(self, event_id, registrant_id):
        event_id = self.event_to_id(event_id)
        if type(registrant_id) is Registrant:
            registrant_id = registrant_id.get_id()
        return Registrant(self.request('get', 'eventspot/events/'+event_id+'/registrants/'+registrant_id))

    def validate_event(self, event):
        if type(event) is dict:
            event = Event(event)
        if type(event) is not Event:
            raise TypeError('invalid type for event: '+type(event).__name__)

        return event

    def validate_fee(self, fee):
        if type(fee) is dict:
            fee = Fee(fee)
        if type(fee) is not Fee:
            raise TypeError('invalid type for fee: '+type(fee).__name__)

        return fee

    def validate_promocode(self, promocode):
        if type(promocode) is dict:
            promocode = Promocode(promocode)
        if type(promocode) is not Promocode:
            raise TypeError('invalid type for promocode: '+type(promocode).__name__)

        return promocode

    def validate_item(self, item):
        if type(item) is dict:
            item = Item(item)
        if type(item) is not Item:
            raise TypeError('invalid type for item: '+type(item).__name__)

        return item

    def validate_attribute(self, attribute):
        if type(attribute) is dict:
            attribute = Attribute(attribute)
        if type(attribute) is not Attribute:
            raise TypeError('invalid type for attribute: '+type(attribute).__name__)

        return attribute

    def event_to_id(self, event_id):
        if type(event_id) is Event:
            event_id = event_id.get_id()
        return event_id

    def fee_to_id(self, fee_id):
        if type(fee_id) is Fee:
            fee_id = fee_id.get_id()
        return fee_id

    def promocode_to_id(self, promocode_id):
        if type(promocode_id) is Promocode:
            promocode_id = promocode_id.get_id()
        return promocode_id

    def item_to_id(self, item_id):
        if type(item_id) is Item:
            item_id = item_id.get_id()
        return item_id

    def attribute_to_id(self, attribute_id):
        if type(attribute_id) is Attribute:
            attribute_id = attribute_id.get_id()
        return attribute_id



from .domain_lambda import *

# Activity_Status is a simple component with no additional logic
# Its fields are all readonly
class Activity_Status(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'contact_count',
            'created_date',
            'error_count',
            'errors',
            'file_name',
            'finish_date',
            'id',
            'start_date',
            'status',
            'type',
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            


from .domain_lambda import *

from .address import Address
from .contact_list import Contact_List
from .email_address import Email_Address

# Contact contains additional logic for fields that contain lists, such as email_addresses, custom_fields, notes, and lists
class Contact(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'custom_fields',
            'prefix_name',
            'first_name',
            'middle_name',
            'last_name',
            'job_title',
            'company_name',
            'home_phone',
            'work_phone',
            'cell_phone',
            'fax',
            'confirmed',
            'created_date',
            'modified_date',
            'source',
            'source_details',
            'id',
            'status'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')

    # The api ignores additional email addresses after the first one
    # This SDK reflects that
    # email_addresses has no remove or clear method as one email_address must always be present
    def get_email_address(self):
        try:
            self['email_addresses'][0]
        except:
            return None
        if type(self['email_addresses'][0]) is dict:
            self['email_addresses'][0] = Email_Address(self['email_addresses'][0])
        return self['email_addresses'][0]
    # In addition to dict and Email_Addresses classes, this method will also accept a string
    # and create a new Email_Address component with the string as the email address
    def set_email_address(self, email_address):
        if type(email_address) is dict:
            email_address = Email_Address(email_address)
        if type(email_address) is Email_Address:
            self['email_addresses'] = [email_address]
        elif type(email_address) is str or unicode:
            new_email = Email_Address()
            new_email.set_email_address(email_address)
            self['email_addresses'] = [new_email]
        else:
            raise TypeError('invalid type for email_address: '+type(email_address).__name__)

    def get_addresses(self):
        try:
            self['addresses']
        except:
            return None
        for i in range(len(self['addresses'])):
            if type(self['addresses'][i]) is not Address:
                self['addresses'][i] = Address(self['addresses'][i])
        return self['addresses']
    # The API currently only supports handling 2 addresses at a time, 
    # one personal and one business address.
    def get_address(self, addr_type):
        try:
            self['addresses']
        except:
            pass
        else:
            for i in range(len(self['addresses'])):
                if type(self['addresses'][i]) is dict:
                    self['addresses'][i] = Address(self['addresses'][i])
                if self['addresses'][i].get_address_type() == addr_type:
                    return self['addresses'][i]
        return None
    def set_address(self, address):
        if type(address) is dict:
            address = Address(address)
        if type(address) is Address:
            if address.get_address_type() not in ['PERSONAL', 'BUSINESS']:
                raise AttributeError('address type must be \'PERSONAL\' or \'BUSINESS\'')
            # Remove address of same type if one already exists
            self.remove_address(address.get_address_type())
            # Initialize field if field doesn't exist
            try:
                self['addresses']
            except:
                self['addresses'] = []
            self['addresses'].append(address)
        else:
            raise TypeError('invalid type for address: '+type(address).__name__)
    def remove_address(self, addr_type = None):
        # Quit method early if there are no addresses
        try:
            self['addresses']
        except:
            return None
        if addr_type:
            for i in range(len(self['addresses'])):
                if self['addresses'][i].get_address_type() == addr_type:
                    self['addresses'].pop(i)
            # Cleanup the dictionary if no more addresses are present
            if len(self['addresses']) is 0:
                self.clear_addresses()
        # if no address type is specified, all addresses are removed
        else:
            self.clear_addresses()

    def clear_addresses(self):
        try:
            self.pop('addresses', None)
        except:
            pass

    # Custom field index does not refer to the position in the list 'custom_fields' 
    # it refers to the numbering of the 'label' of a specific custom field.
    # This is done because custom_fields supports up to 15 fields labelled from 1-15
    # and a user will generally assign a specific field to a certain label.
    def get_custom_field(self, index):
        try:
            self['custom_fields']
        except:
            pass
        else:
            if self['custom_fields']:
                for field in self['custom_fields']:
                    # TODO: Be able to regex match all valid names for custom field labels

                    if str(index) == self.__get_field_name_number(field['name']):
                        return field
        return None

    def set_custom_field(self, index, value):
        self.remove_custom_field(index)
        try:
            self['custom_fields']
        except:
            self['custom_fields'] = []
        field = {'name': 'CustomField'+str(index), 'value': value}
        self['custom_fields'].append(field)

    def remove_custom_field(self, index):
        # Quit method early if there are no custom fields
        try:
            self['custom_fields']
        except:
            return None
        for i in range(len(self['custom_fields'])):
            # TODO: Be able to regex match all valid names for custom field labels

            if str(index) == self.__get_field_name_number(self['custom_fields'][i]['name']):
                self['custom_fields'].pop(i)
                break
        # if no more custom fields are present, cleanup the dict
        if len(self['custom_fields']) == 0:
            self.clear_custom_fields()

    def clear_custom_fields(self):
        try:
            self.pop('custom_fields', None)
        except:
            pass

    def __get_field_name_number(self, name):
        # Extract the name_number from the name field
        # ie for name = '{somestring}123512', name_number = '123512'
        name_number = ''.join([str(char) for char in name if char in [str(d) for d in range(10)]])
        return name_number

    # Methods over the lists list are useful if a user wants to copy lists to another contact 
    def get_lists(self):
        try:
            self['lists']
        except:
            return None
        for i in range(len(self['lists'])):
            if type(self['lists'][i]) is not Contact_List:
                self['lists'][i] = Contact_List(self['lists'][i])
        return self['lists']
    # Does not accept a Result_Set for lists
    def set_lists(self, lists):
        if lists:
            self['lists'] = lists
    def get_list(self, index):
        if type(self['lists'][index]) is dict:
            self['lists'][index] = Contact_List(self['lists'][index])
        return self['lists'][index]
    # Accepts a string as list_id
    def add_list_id(self, list_id):
        try:
            self['lists']
        except:
            self['lists'] = []
        new_list = Contact_List({'id': list_id})
        self['lists'].append(new_list)
    # Accepts a string as list_id
    def add_list(self, contact_list):
        try:
            self['lists']
        except:
            self['lists'] = []
        if type(contact_list) is dict:
            contact_list = Contact_List(contact_list)
        if type(contact_list) is Contact_List:
            self['lists'].append(contact_list)
        else:
            raise TypeError('invalid type for contact_list: '+type(contact_list).__name__)
    def remove_list_id(self, list_id):
        try:
            self['lists']
        except:
            return None
        for i in range(len(self['lists'])):
            if type(self['lists'][i]) is dict:
                self['lists'][i] = Contact_List(self['lists'][i])
            if self['lists'][i].get_id() == list_id:
                self['lists'].pop(i)
                break
    def remove_list(self, index):
        try:
            self['lists']
        except:
            return None
        self['lists'].pop(index)
    # A clear method is added for lists to allow users to easily 
    # reset a contact's lists despite lists being a required field. 
    def remove_all_lists(self):
        try:
            self.pop('lists', None)
        except:
            pass

    # The API only allows max one note per contact
    def get_notes(self):
        try:
            return self['notes']
        except:
            return None
    # note accepts either a dict or string. A string will be formatted into a proper note.
    # There is no validation on an input dict.
    def set_note(self, note):
        try: 
            self['notes']
        except:
            self['notes'] = []
        if type(note) is str:
            self['notes'].append({'note': note})
        elif type(note) is dict:
            self['notes'].append(note)
        else:
            raise TypeError('invalid type for note: '+type(note).__name__)
    def clear_notes(self):
        try:
            self.pop('notes', None)
        except:
            pass




from .domain_lambda import *

from .fee import Fee
from .result_set import Result_Set

# Promocode component
class Promocode(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'code_name',
            'code_type',
            'discount_amount',
            'discount_percent',
            'discount_scope',
            'discount_type',
            'id',
            'is_paused',
            'quantity_available',
            'quantity_total',
            'quantity_used'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')

    # Returns a list of strings
    def get_fee_ids(self):
        try:
            return self['fee_ids']
        except:
            return None

    # Sets the promocode fees to a list of strings or Result_Set of Fees
    def set_fee_ids(self, fee_ids):
        if type(fee_ids) is Result_Set:
            self['fee_ids'] = fee_ids.get_set()
        elif type(fee_ids) is list:
            self['fee_ids'] = fee_ids
        else:
            raise TypeError('invalid type for fee_ids: '+type(fee_ids).__name__)
        for i in range(len(self['fee_ids'])):
            if type(self['fee_ids'][i]) is Fee:
                self['fee_ids'][i] = self['fee_ids'][i].get_id()

    # Deletes a fee_id by value
    def delete_fee_by_id(self, fee_id):
        try:
            self['fee_ids']
        except:
            return None
        for i in range(len(self['fee_ids'])):
            if self['fee_ids'][i] == fee_id:
                self['fee_ids'].pop(i)
                break

    # Deletes a fee_id by index
    def delete_fee(self, index):
        try:
            self['fee_ids']
        except:
            return None
        self['fee_ids'].pop(index)

    # Clears all fees
    def clear_fee_ids(self):
        try:
            self.pop('fee_ids', None)
        except:
            pass




from .baseservice import BaseService

from .components import Billing_Plan, Partner_Account, Result_Set

from datetime import datetime

class PartnerService(BaseService):
    # Partner services endpoints
    def get_partner_accounts(self, limit = 50):
        queries = {}
        queries['limit'] = limit
        
        response = self.request('get', 'partner/accounts', params = queries)
        try:
            response['results']
        except:
            return response
        # Convert dict to Components
        for i in range(len(response['results'])):
            response['results'][i] = Partner_Account(response['results'][i])
        return Result_Set(response)

    def post_partner_accounts(self, account, token = False):
        return Partner_Account(self.request('post', 'partner/accounts', params = {'oauth2_token_provided': token}, body = account))
    
    def get_partner_billing_plan(self, partner_id):
        partner_id = self.partner_to_id(partner_id)
        return Billing_Plan(self.request('get', 'partner/accounts/'+partner_id+'/plan'))
    
    def update_partner_billing_plan(self, partner_id, plan):
        partner_id = self.partner_to_id(partner_id)
        return Billing_Plan(self.request('put', 'partner/accounts/'+partner_id+'/plan', body = plan))

    def cancel_account(self, partner_id, cancel_reason = None):
        body = {'status': 'Cancelled'}
        if cancel_reason:
            body['cancel_reason'] = cancel_reason
        date = datetime.now().isoformat().split('T', 1)[0]
        body['cancel_date'] = date

        partner_id = self.partner_to_id(partner_id)

        return self.request('put', 'partner/accounts/'+partner_id+'/status', body = body)

    def partner_to_id(self, partner_id):
        if type(partner_id) is Partner_Account:
            partner_id = partner_id.get_id()
        return partner_id





from .baseservice import BaseService

from .components import Multipart_File, Mylibrary_File, Mylibrary_Folder, Result_Set

# MyLibrary Endpoints
class MyLibraryService(BaseService):

    # Files endpoints
    def get_files(self, limit = 50, sort_by = 'CREATED_DATE_DESC', source = 'ALL', file_type = 'ALL'):
        queries = {}
        queries['limit'] = limit
        queries['sort_by'] = sort_by
        queries['source'] = source
        queries['type'] = file_type

        response = self.request('get', 'library/files', params = queries)
        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Mylibrary_File(response['results'][i])
        return Result_Set(response)

    # Accepts a Multipart_File or a dict
    def post_files(self, multipart_file):
        attributes = [
            'data',
            'description',
            'file_name',
            'file_type',
            'folder_id',
            'source'
        ]
        form_file, form_data = self.prepare_multipart_body(multipart_file, attributes)

        return self.request('post', 'library/files', form_file = form_file, form_data = form_data)

    # Accepts a list of file_ids or Mylibrary_Files
    # There is no component for file upload status, so a list of dicts will be returned.
    def get_files_upload_status(self, file_ids):
        ids = ''
        for file_id in file_ids:
            if type(file_id) is Mylibrary_File:
                ids = ids + file_id.get_id() + ','
            elif type(file_id) is str:
                ids = ids + file_id + ','
            else:
                raise TypeError('Unknown type for file_id: '+type(file_id).__name__)

        # strip out the last comma
        ids = ids[:-1]

        return self.request('get', 'library/files/uploadstatus/' + ids)

    def get_file(self, file_id):
        file_id = self.file_to_id(file_id)
        return Mylibrary_File(self.request('get', 'library/files/' + file_id))

    def update_file(self, mylibrary_file, include_payload = True):
        params = {'include_payload' : include_payload}
        mylibrary_file = self.validate_file(mylibrary_file)

        return Mylibrary_File(self.request('put', 'library/files/' + mylibrary_file.get_id(), body = mylibrary_file, params = params))

    def delete_file(self, file_id):
        file_id = self.file_to_id(file_id)
        return self.request('delete', 'library/files/' + file_id)

    # Folders endpoints
    def get_folders(self, limit = 50, sort_by = 'CREATED_DATE_DESC'):
        queries = {}
        queries['limit'] = limit
        queries['sort_by'] = sort_by

        response = self.request('get', 'library/folders', params = queries)
        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Mylibrary_Folder(response['results'][i])
        return Result_Set(response)

    def post_folders(self, folder):
        folder = self.validate_folder(folder)
        return Mylibrary_Folder(self.request('post', 'library/folders', body = folder))

    def get_folder(self, folder_id):
        folder_id = self.folder_to_id(folder_id)
        return Mylibrary_Folder(self.request('get', 'library/folders/' + folder_id))

    def update_folder(self, folder):
        folder = self.validate_folder(folder)
        return Mylibrary_Folder(self.request('put', 'library/folders/' + folder.get_id(), body = folder))

    def delete_folder(self, folder_id):
        folder_id = self.folder_to_id(folder_id)
        return self.request('delete', 'library/folders/' + folder_id)

    def get_folder_files(self, folder_id, limit = 50):
        queries = {}
        queries['limit'] = limit

        folder_id = self.folder_to_id(folder_id)
        return self.request('get', 'library/folders/' + folder_id + '/files', params = queries)
        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Mylibrary_Folder(response['results'][i])
        return Result_Set(response)

    # Accepts a list of file_ids or Mylibrary_Files
    def move_folder_files(self, folder_id, file_ids):
        folder_id = self.folder_to_id(folder_id)
        ids = []
        for file_id in file_ids:
            if type(file_id) is Mylibrary_File:
                ids.append(file_id.get_id())
            elif type(file_id) is str:
                ids.append(file_id)
            else:
                raise TypeError('Unknown type for file_id: '+type(file_id).__name__)

        return self.request('put', 'library/folders/' + folder_id + '/files', body = ids)

    # Trash endpoints
    def get_trash(self, limit = 50, sort_by = 'CREATED_DATE_DESC', file_type = 'ALL'):
        queries = {}
        queries['limit'] = limit
        queries['sort_by'] = sort_by
        queries['type'] = file_type

        response = self.request('get', 'library/folders/trash/files', params = queries)
        try:
            response['results']
        except:
            return response
        for i in range(len(response['results'])):
            response['results'][i] = Mylibrary_File(response['results'][i])
        return Result_Set(response)
    def delete_trash(self):
        return self.request('delete', 'library/folders/trash/files')

    # Mylibrary info endpoint
    # Returns a dict as there is for component for mylibrary info
    def get_mylibrary_info(self):
        return self.request('get', 'library/info')

    def validate_file(self, mylibrary_file):
        if type(mylibrary_file) is dict:
            mylibrary_file = Mylibrary_File(mylibrary_file)
        if type(mylibrary_file) is not Mylibrary_File:
            raise TypeError('invalid type for Mylibrary file: '+type(mylibrary_file).__name__)

        return mylibrary_file

    def validate_folder(self, folder):
        if type(folder) is dict:
            folder = Mylibrary_Folder(folder)
        if type(folder) is not Mylibrary_Folder:
            raise TypeError('invalid type for Mylibrary folder: '+type(folder).__name__)

        return folder

    def file_to_id(self, file_id):
        if type(file_id) is Mylibrary_File:
            file_id = file_id.get_id()
        return file_id

    def folder_to_id(self, folder_id):
        if type(folder_id) is Mylibrary_Folder:
            folder_id = folder_id.get_id()
        return folder_id









from .domain_lambda import *

# Billing_Plan is a simple component with no additional logic
class Billing_Plan(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Current tiers is returned as a dict
        attributes = [
            'billing_day_of_month',
            'current_tiers',
            'plan_level',
            'plan_type',
            'status'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')


from .domain_lambda import *

# Activity_Report is a simple component with no additional logic
class Activity_Report(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'activity_type',
            'bounce_code',
            'bounce_date',
            'bounce_description',
            'bounce_message',
            'bounces',
            'campaign_id',
            'click_date',
            'clicks',
            'contact_id',
            'email_address',
            'forward_date',
            'forwards',
            'link_id',
            'link_uri',
            'open_date',
            'opens',
            'send_date',
            'unsubscribe_date',
            'unsubscribe_reason',
            'unsubscribe_source'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            


from .domain_lambda import *

# Event Registrant component
class Registrant(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # All attributes are read-only
        # Array and object attriutes for the Registrant component 
        # will not have a specialized component for them
        attributes = [
            'attendance_status',
            'email',
            'fees',
            'guest_info',
            'guest_secton',
            'guests',
            'first_name',
            'guest_count',
            'id',
            'items',
            'last_name',
            'order',
            'payment_status',
            'payment_summary',
            'promo_code_info',
            'registration_date',
            'registration_status',
            'sections',
            'ticket_id',
            'updated_date'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')




from .baseservice import BaseService

from .components import Account_Info, Email_Address

class AccountService(BaseService):
    # Account services endpoints
    def get_account_info(self):
        return Account_Info(self.request('get', 'account/info'))
        
    def update_account_info(self, account_info):
        account_info = self.validate_account_info(account_info)
        return Account_Info(self.request('put', 'account/info', body = account_info))

    # Email addresses are returned as standard python objects rather than a Result_Set containing Email_Address components
    def get_verified_email_addresses(self, status = 'ALL'):
        return self.request('get', 'account/verifiedemailaddresses', params = {'status': status})

    def post_verified_email_addresses(self, email):
        if type(email) is Email_Address:
            email = email.get_email_address()
        return self.request('post', 'account/verifiedemailaddresses', body = [{'email_address': email}])

    def validate_account_info(self, account_info):
        if type(account_info) is dict:
            account_info = Account_Info(account_info)
        if type(account_info) is not Account_Info:
            raise TypeError('invalid type for account_info: '+type(account_info).__name__)
            
        return account_info




##
# A simple python program utilizing the constant contact API
##

from .accountservice import AccountService
from .activityservice import ActivityService
from .contactservice import ContactService
from .emailmarketingservice import EmailMarketing
from .eventspotservice import EventSpotService
from .listservice import ListService
from .mylibraryservice import MyLibraryService
from .partnerservice import PartnerService

class ConstantContact:
    # The base url is the Constant Contact v2 API
    def __init__(self, client_id, token, base_url = 'https://api.constantcontact.com/v2'):
        self.__init_services__(client_id, token, base_url)

    def __init_services__(self, client_id, token, base_url):
        # Initialize services
        self.account_service = AccountService(client_id, token, base_url)
        self.emailmarketing_service = EmailMarketing(client_id, token, base_url)
        self.mylibrary_service = MyLibraryService(client_id, token, base_url)
        self.activity_service = ActivityService(client_id, token, base_url)
        self.contact_service = ContactService(client_id, token, base_url)
        self.list_service = ListService(client_id, token, base_url)
        self.partner_service = PartnerService(client_id, token, base_url)
        self.eventspot_service = EventSpotService(client_id, token, base_url)

    # Token info endpoint
    def get_token_info(self):
        return self.account_service.token_info()

    # Account services endpoints
    def get_account_info(self):
        return self.account_service.get_account_info()
    def update_account_info(self, account_info):
        return self.account_service.update_account_info(account_info)
    def get_verified_email_addresses(self, status = 'ALL'):
        return self.account_service.get_verified_email_addresses(status)
    def post_verified_email_addresses(self, email):
        return self.account_service.post_verified_email_addresses(email)

    # Email Marketing Endpoints
    def get_campaigns(self, limit = 50, modified_since = None, status = 'ALL'):
        return self.emailmarketing_service.get_campaigns(limit, modified_since, status)
    def post_campaigns(self, campaign):
        return self.emailmarketing_service.post_campaigns(campaign)
    def get_campaign(self, campaign_id, updateSummary = False):
        return self.emailmarketing_service.get_campaign(campaign_id, updateSummary)
    def update_campaign(self, campaign):
        return self.emailmarketing_service.update_campaign(campaign)
    def delete_campaign(self, campaign_id):
        return self.emailmarketing_service.delete_campaign(campaign_id)

    def test_send_campaign(self, campaign_id, email_addresses, format, personal_message):
        return self.emailmarketing_service.test_send_campaign(campaign_id, email_addresses, format, personal_message)
    def preview_campaign(self, campaign_id):
        return self.emailmarketing_service.preview_campaign(campaign_id)

    def get_campaign_schedules(self, campaign_id):
        return self.emailmarketing_service.get_campaign_schedules(campaign_id)
    def post_campaign_schedules(self, campaign_id, schedule):
        return self.emailmarketing_service.post_campaign_schedules(campaign_id, schedule)
    def get_campaign_schedule(self, campaign_id, schedule_id):
        return self.emailmarketing_service.get_campaign_schedule(campaign_id, schedule_id)
    def update_campaign_schedule(self, campaign_id, schedule):
        return self.emailmarketing_service.update_campaign_schedule(campaign_id, schedule)
    def delete_campaign_schedule(self, campaign_id, schedule_id):
        return self.emailmarketing_service.delete_campaign_schedule(campaign_id, schedule_id)

    def get_campaign_summary_report(self, campaign_id, updateSummary = False):
        return self.emailmarketing_service.get_campaign_summary_report(campaign_id, updateSummary)
    def get_campaign_bounces(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_bounces(campaign_id, created_since, limit)
    def get_campaign_clicks(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_clicks(campaign_id, created_since, limit)
    def get_campaign_forwards(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_forwards(campaign_id, created_since, limit)
    def get_campaign_opens(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_opens(campaign_id, created_since, limit)
    def get_campaign_sends(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_sends(campaign_id, created_since, limit)
    def get_campaign_unsubscribes(self, campaign_id, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_unsubscribes(campaign_id, created_since, limit)
    def get_campaign_clicks_by_link_id(self, campaign_id, link_id = None, created_since = None, limit = 500):
        return self.emailmarketing_service.get_campaign_clicks_by_link_id(campaign_id, link_id, created_since, limit)

    # My Library Endpoints
    def get_files(self, limit = 50, sort_by = 'CREATED_DATE_DESC', source = 'ALL', file_type = 'ALL'):
        return self.mylibrary_service.get_files(limit, sort_by, source, file_type)
    def post_files(self, multipart_file):
        return self.mylibrary_service.post_files(multipart_file)
    def get_files_upload_status(self, file_ids):
        return self.mylibrary_service.get_files_upload_status(file_ids)
    def get_file(self, file_id):
        return self.mylibrary_service.get_file( file_id)
    def update_file(self, mylibrary_file, include_payload = True):
        return self.mylibrary_service.update_file(mylibrary_file, include_payload)
    def delete_file(self, file_id):
        return self.mylibrary_service.delete_file(file_id)

    def get_folders(self, limit = 50, sort_by = 'CREATED_DATE_DESC'):
        return self.mylibrary_service.get_folders(limit, sort_by)
    def post_folders(self, folder):
        return self.mylibrary_service.post_folders(folder)
    def get_folder(self, folder_id):
        return self.mylibrary_service.get_folder(folder_id)
    def update_folder(self, folder):
        return self.mylibrary_service.update_folder(folder)
    def delete_folder(self, folder_id):
        return self.mylibrary_service.delete_folder(folder_id)
    def get_folder_files(self, folder_id, limit = 50):
        return self.mylibrary_service.get_folder_files(folder_id, limit)
    def move_folder_files(self, folder_id, file_ids):
        return self.mylibrary_service.move_folder_files(folder_id, file_ids)

    def get_trash(self, limit = 50, sort_by = 'CREATED_DATE_DESC', file_type = 'ALL'):
        return self.mylibrary_service.get_trash(limit, sort_by, file_type)
    def delete_trash(self):
        return self.mylibrary_service.delete_trash()
    def get_mylibrary_info(self):
        return self.mylibrary_service.get_mylibrary_info()

    # Activity services endpoints
    def import_contacts(self, import_data, lists, columns):
        return self.activity_service.import_contacts(import_data, lists, columns)
    def remove_contacts(self, import_data, lists):
        return self.activity_service.remove_contacts(import_data, lists)
    def clear_contact_lists(self, lists):
        return self.activity_service.clear_contact_lists(lists)
    def export_contacts(self, columns, lists, sort_by = 'EMAIL_ADDRESS', file_type = 'CSV', export_added_by = False, export_date_added = False):
        return self.activity_service.export_contacts(columns, lists, sort_by, file_type, export_added_by, export_date_added)
    def get_export_file(self, file_name):
        return self.activity_service.get_export_file(file_name)
    def multipart_import_contacts(self, multipart_activity):
        return self.activity_service.multipart_import_contacts(multipart_activity)
    def multipart_remove_contacts(self, multipart_activity):
        return self.activity_service.multipart_remove_contacts(multipart_activity)
    def activities_status_report(self, type_query = None, status = 'ALL'):
        return self.activity_service.activities_status_report(type_query, status)
    def individual_activity_status(self, activity_id):
        return self.activity_service.individual_activity_status(activity_id)

    # Contacts services endpoints
    def get_contacts(self, email = None, limit = 50, modified_since = None, status = 'ALL'):
        return self.contact_service.get_contacts(email, limit, modified_since, status)
    def post_contacts(self, contact, action_by = 'ACTION_BY_OWNER'):
        return self.contact_service.post_contacts(contact, action_by)
    def get_contact(self, contact_id):
        return self.contact_service.get_contact(contact_id)
    def update_contact(self, contact, action_by = 'ACTION_BY_OWNER'):
        return self.contact_service.update_contact(contact, action_by)
    def delete_contact(self, contact_id):
        return self.contact_service.delete_contact(contact_id)

    # Contacts tracking endpoints
    def get_contact_tracking(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_tracking(contact_id, created_since, limit)
    def get_contact_tracking_summary(self, contact_id):
        return self.contact_service.get_contact_tracking_summary(contact_id)
    def get_contact_bounces(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_bounces(contact_id, created_since, limit)
    def get_contact_clicks(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_clicks(contact_id, created_since, limit)
    def get_contact_forwards(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_forwards(contact_id, created_since, limit)
    def get_contact_opens(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_opens(contact_id, created_since, limit)
    def get_contact_sends(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_sends(contact_id, created_since, limit)
    def get_contact_unsubscribes(self, contact_id, created_since = None, limit = 500):
        return self.contact_service.get_contact_unsubscribes(contact_id, created_since, limit)
    def get_contact_report_by_campaign(self, contact_id):
        return self.contact_service.get_contact_report_by_campaign(contact_id)
    
    # Contact List services endpoints
    def get_lists(self, modified_since = None):
        return self.list_service.get_lists(modified_since)
    def post_lists(self, list_name, status = 'ACTIVE'):
        return self.list_service.post_lists(list_name, status)
    def get_list(self, list_id):
        return self.list_service.get_list(list_id)
    def update_list(self, contact_list):
        return self.list_service.update_list(contact_list)
    def delete_list(self, list_id):
        return self.list_service.delete_list(list_id)
    def get_list_contacts(self, list_id, limit = 50, modified_since = None):
        return self.list_service.get_list_contacts(list_id, limit, modified_since)

    # Partner services endpoints
    def get_events(self, limit = 50):
        return self.eventspot_service.get_events(limit)
    def post_events(self, event):
        return self.eventspot_service.post_events(event)
    def get_event(self, event_id):
        return self.eventspot_service.get_event(event_id)
    def update_event(self, event):
        return self.eventspot_service.update_event(event)
    def patch_event(self, event_id, value):
        return self.eventspot_service.patch_event(event_id, value)

    def get_event_promocodes(self, event_id):
        return self.eventspot_service.get_event_promocodes(event_id)
    def post_event_promocodes(self, event_id, promocode):
        return self.eventspot_service.post_event_promocodes(event_id, promocode)
    def get_event_promocode(self, event_id, promocode_id):
        return self.eventspot_service.get_event_promocode(event_id, promocode_id)
    def update_event_promocode(self, event_id, promocode):
        return self.eventspot_service.update_event_promocode(event_id, promocode)
    def delete_event_promocode(self, event_id, promocode_id):
        return self.eventspot_service.delete_event_promocode(event_id, promocode_id)

    def get_event_fees(self, event_id):
        return self.eventspot_service.get_event_fees(event_id)
    def post_event_fees(self, event_id, fee):
        return self.eventspot_service.post_event_fees(event_id, fee)
    def get_event_fee(self, event_id, fee_id):
        return self.eventspot_service.get_event_fee(event_id, fee_id)
    def update_event_fee(self, event_id, fee):
        return self.eventspot_service.update_event_fee(event_id, fee)
    def delete_event_fee(self, event_id, fee_id):
        return self.eventspot_service.delete_event_fee(event_id, fee_id)

    def get_event_items(self, event_id):
        return self.eventspot_service.get_event_items(event_id)
    def post_event_items(self, event_id, item):
        return self.eventspot_service.post_event_items(event_id, item)
    def get_event_item(self, event_id, item_id):
        return self.eventspot_service.get_event_item(event_id, item_id)
    def update_event_item(self, event_id, item):
        return self.eventspot_service.update_event_item(event_id, item)
    def delete_event_item(self, event_id, item_id):
        return self.eventspot_service.delete_event_item(event_id, item_id)

    def get_event_item_attributes(self, event_id, item_id):
        return self.eventspot_service.get_event_item_attributes(event_id, item_id)
    def post_event_item_attributes(self, event_id, item_id, attribute):
        return self.eventspot_service.post_event_item_attributes(event_id, item_id, attribute)
    def get_event_item_attribute(self, event_id, item_id, attribute_id):
        return self.eventspot_service.get_event_item_attribute(event_id, item_id, attribute_id)
    def update_event_item_attribute(self, event_id, item_id, attribute):
        return self.eventspot_service.update_event_item_attribute(event_id, item_id, attribute)
    def delete_event_item_attribute(self, event_id, item_id, attribute_id):
        return self.eventspot_service.delete_event_item_attribute(event_id, item_id, attribute_id)

    def get_event_registrants(self, event_id, limit = 50):
        return self.eventspot_service.get_event_registrants(event_id, limit)
    def get_event_registrant(self, event_id, registrant_id):
        return self.eventspot_service.get_event_registrant(event_id, registrant_id)

    # Partner services endpoints
    def get_partner_accounts(self, limit = 50):
        return self.partner_service.get_partner_accounts(limit)
    def post_partner_accounts(self, account, token = False):
        return self.partner_service.post_partner_accounts(account, token)
    def get_partner_billing_plan(self, partner_id):
        return self.partner_service.get_partner_billing_plan(account_id)
    def update_partner_billing_plan(self, partner_id, plan):
        return self.partner_service.update_partner_billing_plan(account_id, plan)
    def cancel_account(self, partner_id):
        return self.partner_service.cancel_account(account_id)

    # Paginated results
    def next_page(self, page_uri):
        return self.account_service.next_page(page_uri)



from .domain_lambda import *

# Item Attribute component
class Attribute(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'id',
            'name',
            'quantity_available',
            'quantity_total'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')



from .domain_lambda import *

from .attribute import Attribute
from .result_set import Result_Set

# Item component
class Item(dict):

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'default_quantity_available',
            'default_quantity_total',
            'description',
            'id',
            'name',
            'per_registrant_limit',
            'price',
            'show_quantity_available'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')

    def get_attributes(self):
        try:
            self['attributes']
        except:
            return None
        for i in range(len(self['attributes'])):
            self['attributes'][i] = Attribute(self['attributes'][i])
        return self['attributes']

    def set_attributes(self, attributes):
        if type(attributes) is Result_Set:
            self['attributes'] = attributes.get_set()
        elif type(attributes) is list:
            self['attributes'] = attributes
        else:
            raise TypeError('invalid type for attributes: '+type(attributes).__name__)

    # Deletes an attribute by id
    def delete_attribute_by_id(self, attribute_id):
        try:
            self['attributes']
        except:
            return None
        for i in range(len(self['attributes'])):
            try:
                if self['attributes'][i]['id'] == attribute_id:
                    self['attributes'].pop(i)
                    break
            except:
                pass

    # Deletes an attribute_id by index
    def delete_attribute(self, index):
        try:
            self['attributes']
        except:
            return None
        self['attributes'].pop(index)

    # Clears all attributes
    def clear_attributes(self):
        try:
            self.pop('attributes', None)
        except:
            pass




"""
Constant Contact Python SDK for v2 API.
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

from constantcontact import __version__

# Get the long description from the relevant file
try:
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, 'readme.md'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = 'Description can be found at: https://github.com/magellantoo/ctct-python-sdk'

setup(
    name='constantcontact',
    # Version found in /constantcontact/baseservice.py
    version=__version__,
    description='A Python SDK for the Constant Contact v2 API',
    long_description=long_description,
    url='http://developer.constantcontact.com/libraries/libraries.html',
    author='Constant Contact Inc.',
    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],

    keywords='constantcontact sdk',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    #package_dir={'python-sdk': 'python-sdk'},

    install_requires=['requests'],
)



from .domain_lambda import *

class Email_Address(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        attributes = [
            'id',
            'status',
            'confirm_status',
            'opt_in_source',
            'opt_out_source',
            'opt_in_date',
            'opt_out_date',
            'email_address',
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')



from .domain_lambda import *

# Mylibrary Folder component
class Mylibrary_Folder(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # Children is stored as an array of strings
        attributes = [
            'children',
            'created_date',
            'id',
            'item_count',
            'level',
            'modified_date',
            'name',
            'parent_id'
        ]
        
        for name in attributes:
            exec('self.get_'+name+' = make_getter(self, \''+name+'\')')
            exec('self.set_'+name+' = make_setter(self, \''+name+'\')')
            exec('self.delete_'+name+' = make_deleter(self, \''+name+'\')')
            